/*
 * Process-related convenience functions.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>

#include <stdio.h>
#include <stdlib.h>

#include "process_creator_client.h"

#include "file.h"
#include "process.h"
#include "notifier.h"



/* Create a new process object. */

process_t *process_new()
{
  process_t *process = (process_t *) malloc(sizeof(process_t));

  if (process == NULL)
    return NULL;

  process_init(process);
  return process;
}

/* Initialise the given process structure. */

void process_init(process_t *process)
{
  process->ref = L4_INVALID_CAP;

  /* Initialise the notifiable section of the structure. */

  process->notifiable.notifications = 0;
  process->notifiable.base = (notifiable_base_t *) process;
}

/* Start a process using the given file as payload.
   NOTE: This does not yet communicate arguments or obtain input/output
         pipes. */

long process_start(process_t *process, file_t *file, process_notifier_t *notifier)
{
  l4_cap_idx_t server = l4re_env_get_cap("pserver");

  if (l4_is_invalid_cap(server))
    return -L4_ENOMEM;

  /* Obtain a notification endpoint. */

  l4_cap_idx_t endpoint;
  long err = process_notify_get_endpoint(process, notifier, &endpoint);

  if (err)
    return err;

  /* Obtain a client for the process creator. */

  client_ProcessCreator creator(server);

  /* Start the process, supplying the given endpoint. */

  return creator.start(file->ref, endpoint);
}



/* NOTE: Much of the code below could be unified with the file-related
         notification code. */

/* Opaque notifier type for process_notifier_t. */

struct process_notifier
{
  ObjectNotifier *obj;
};

/* Conversion to the generic notification types. */

notifiable_t *process_notifiable(process_t *process)
{
  return &process->notifiable;
}

/* Return the notification flags for a process. */

notify_flags_t process_notifications(process_t *process)
{
  return process->notifiable.notifications;
}

/* Return the notification values for a process. */

notify_values_t process_notification_values(process_t *process)
{
  return process->notifiable.values;
}

/* Close a notifier object. */

void process_notify_close(process_notifier_t *notifier)
{
  delete notifier->obj;
  delete notifier;
}

/* Obtain a local notifier object. */

process_notifier_t *process_notify_local()
{
  process_notifier_t *notifier = new process_notifier_t;

  notifier->obj = notifier_get_local_notifier();
  return notifier;
}

/* Obtain the task-wide notifier object. */

process_notifier_t *process_notify_task()
{
  process_notifier_t *notifier = new process_notifier_t;

  notifier->obj = notifier_get_task_notifier();
  return notifier;
}

/* Subscribe to notification events on a process. */

long process_notify_get_endpoint(process_t *process, process_notifier_t *notifier, l4_cap_idx_t *endpoint)
{
  return notifier->obj->get_endpoint(process_notifiable(process), endpoint, true);
}

/* Unsubscribe from notification events on a process. */

long process_notify_remove_endpoint(process_t *process, process_notifier_t *notifier, l4_cap_idx_t endpoint)
{
  return notifier->obj->remove_endpoint(process_notifiable(process), endpoint);
}

/* Wait for a notification event on a process. */

long process_notify_wait_process(process_t *process, process_notifier_t *notifier)
{
  SpecificObjectNotifier *specific_notifier = dynamic_cast<SpecificObjectNotifier *>(notifier->obj);
  long err = specific_notifier->wait_object(process_notifiable(process));

  /* Unsubscribe if a closure notification has been received. */

  //if (!err && (process->notifiable.notifications & NOTIFY_PEER_CLOSED))
    //process_notify_unsubscribe(process, notifier);

  return err;
}

/* Wait for notification events on processes. */

long process_notify_wait_processes(process_t **process, process_notifier_t *notifier)
{
  GeneralObjectNotifier *general_notifier = dynamic_cast<GeneralObjectNotifier *>(notifier->obj);
  notifiable_t *notifiable;
  long err = general_notifier->wait(&notifiable);

  *process = (process_t *) notifiable->base;

  /* Unsubscribe if a closure notification has been received. */

  //if (!err && ((*file)->notifiable.notifications & NOTIFY_PEER_CLOSED))
    //file_notify_unsubscribe(*file, notifier);

  return err;
}

// vim: tabstop=2 expandtab shiftwidth=2
